# -*- coding: utf-8 -*-
"""
File name:
snow_scrip_main_vXXX.py

Description:
Functions for extracting snow cover and snow depth from NEON snow and
ground phenology cameras.  This code is written in a way that is conducive to 
cutting and pasting code chunks into an IPython console or notebook. It is not
written to run the whole thing at once from a Bash or DOS shell.

This is experimental code, for proof of concept. It requires a lot of context 
specific fine tuning.  The inline comments are descriptive but not overly 
detailed. Getting this code to work on anything but the original set of test 
images may be challenging.  Hope you find it useful.

Created by:
Tim Meehan
Staff Scientist
National Ecological Observatory Network

Change log:
Version 002 created on 3 March 2016. Added lots of comments.
"""


#################################  load libraries  ############################

"""
First, we need to import a bunch of python libraries.
"""

# import
import os # access to operating system commands
import glob # make lists of files
import numpy as np # array computing routines
import itertools # combinometric functions
import datetime # massaging dates and times
from scipy import ndimage as nd # image processing
from skimage import exposure, filter # image processing
import matplotlib.pylab as plt # plotting
import matplotlib.dates as mpd # plotting dates and times
import matplotlib.gridspec as gridspec # arranging plots
from pylab import savefig # saving plots
import matplotlib.animation as animation # making animations from images
from scipy import misc
import pandas as pd # working with spreadsheet-like objects


##################################  define functions  #########################

"""
Next we create some functions that we can reuse for repetitive operations. One
gets snow depths from staff gauges.  Another gets percent cover from reference
areas of ground.  A third summarizes output from the first two.
"""

# get snow depth from a ruler
def get_ruler_depths(flist, rotangle = -10.0, tlx = 400, tly = 400,
               delx = 10, dely = 260, x0 = 5, y0 = 5, x1 = 5, y1 = 250,
               cm = 100.0, pix = 150.0, cut = 8, rep_size = 4, sd_thresh = 10):
    """
    This function extracts snow depths from a series of images. Inputs include
    a list of image files (flist), an angle (rotangle) for rotation of image
    so that snow ruler is close to vertical, the top left corner x-coordinate
    of the ruler after rotation (tlx), the top left corner y-coordinate of the
    ruler after rotation (tly), the width in pixels of the ruler after
    rotation (delx), the length in pixels of the ruler after rotation (dely),
    the starting (x0, y0) and ending (x1, y1) coordinates for the line transect
    along the ruler after rotation and subsetting, a number of centimeters
    along the transect where the number of pixels is calculated (cm), the
    number of pixels counted over the number of centimeters listed above (pix),
    a cutoff number of white pixels before a white space is considered as
    snow (cut), the number of consecutive images in a temporal block to be 
    considered as replicates (rep_size), and the threshold in cm beyond which 
    we set a quality flag indication that there was too much light pollution 
    to get a good estimate (sd_thresh).

    Outputs comprise numpy arrays representing time series of mean snow 
    depths in cm, standard deviation of snow depths in cm, and a quality flags
    per temporal block specified by rep size.
    """
    # create lists of depths and times
    depth_list = []
    time_list = []
    for f in flist:
        print f
        # get time stamp information
        tstr1 =  f.split("_")
        yr = int(tstr1[1])
        mo = int(tstr1[2])
        da = int(tstr1[3])
        ts = tstr1[4].split(".")[0]
        hr = int(ts[0:2])
        mi = int(ts[2:4])
        se = int(ts[4:6])
        dt = datetime.datetime(yr,mo,da,hr,mi,se)
        # stash the date object in a list
        time_list.append(dt)
        # get original image file and make it grayscale (0-255)
        im1 = nd.imread(f, flatten=1)
        # rotate image so ROI is vertical, parallel to image borders
        im2 = nd.rotate(im1, rotangle)
        # subset image down to a snow gauge
        im3 = im2[tly:tly+dely, tlx:tlx+delx]
        # rescale image exposure to deal with variation in lighting
        im4 = exposure.rescale_intensity(im3)
        # make binary for counting
        val = filter.threshold_otsu(im4)
        im5 = im4 > val
        im5.astype(int)
        # set linear transect over binary gauge image
        length = int(np.hypot(x1-x0, y1-y0))
        x, y = np.linspace(x0, x1, length), np.linspace(y0, y1, length)
        # extract the binary values along the transect
        zi1 = im5[y.astype(np.int), x.astype(np.int)].astype(int)
        # calculate the number of consecutive ones (snow pixels) along transect
        reps1 = [sum( 1 for _ in group ) for key, group in
        itertools.groupby(zi1) if key ]
        # this is where short runs of bright pixels are zeroed out using
        # the cut parameter. this gets rid of bright spots on rulers
        # in between numbers.
        for i in range(0,len(reps1)):
            if reps1[i] >= cut:
                reps1[i] = reps1[i]
            else:
                reps1[i] = 0
        # now sum the larger consectutive strings of snow pixels together
        snow_pix = np.array(reps1).sum()
        # conversion from number of pixels to snow depth in cm
        conv_factor = cm / pix
        depth_cm = (snow_pix * conv_factor); depth_cm
        # stash result in the list
        depth_list.append(depth_cm)
    # turn lists to numpy arrays
    depth_list = np.array((depth_list))
    time_list = np.array((time_list))
    # calculate replicate means and standard deviations for the temporal blocks
    # specified by rep size
    starts = range(0, len(depth_list), rep_size)
    rep_means = []
    rep_std = []
    rep_std_flag = []
    rep_time = []
    for start in starts:
        # create start and end indices
        block_start = start
        block_end = start + rep_size
        # get mean for block
        mean_i = np.nanmean(depth_list[block_start : block_end])
        # get sd for block
        sd_i = np.nanstd(depth_list[block_start : block_end])
        # set sd flag if necessary, where 1 indicates a faulty value
        sd_flag_i = 0
        if sd_i < sd_thresh:
            sd_flag_i = 0
        else:
            sd_flag_i = 1
        # get mean time of block
        med_index = block_start
        time_i = time_list[med_index]
        # append to lists
        rep_means.append(mean_i)
        rep_std.append(sd_i)
        rep_std_flag.append(sd_flag_i)
        rep_time.append(time_i)
    # bundle results
    out_list = [rep_means, rep_std, rep_std_flag, rep_time]
    return out_list

# get snow cover from a patch of ground
def get_snow_cover(flist, tlx = 250, tly = 350,
               delx = 150, dely = 100, sum_snow = 12955,
               sum_nosnow = 4853, rep_size = 4, sd_thresh = 20):
    """
    This function extracts percent snow cover from a series of images. Inputs
    include a list of image files (flist), the top left corner x-coordinate
    of the reference patch (tlx), the top left corner y-coordinate of the
    reference patch (tly), the width in pixels of the reference patch (delx),
    the length in pixels of reference patch (dely), the number of bright pixels
    in an image with complete snow cover (sum_snow), the number of bright
    pixels in an image with no snow cover (sum_nosnow), the number of 
    consecutive images in a temporal block to be considered as replicates 
    (rep_size), and the threshold in percent beyond which we set a quality 
    flag indicating that there was too much light pollution to get a good 
    estimate (sd_thresh).

    Outputs comprise numpy arrays representing time series of mean percent snow 
    cover, standard deviation of snow cover, and quality flags per temporal 
    block specified by rep size.
    """
    # create list of depths
    depth_list = []
    time_list = []
    for f in flist:
        print f
        # get time stamps
        tstr1 =  f.split("_")
        yr = int(tstr1[1])
        mo = int(tstr1[2])
        da = int(tstr1[3])
        ts = tstr1[4].split(".")[0]
        hr = int(ts[0:2])
        mi = int(ts[2:4])
        se = int(ts[4:6])
        dt = datetime.datetime(yr,mo,da,hr,mi,se)
        # stash the date time object
        time_list.append(dt)
        # get the original input file and make it grayscale (0-255)
        im1 = nd.imread(f, flatten=1)
        im2 = im1 # this is a hack, ignore it
        # subset image to a reference area of ground that has something
        # dark that won't be covered by snow
        im3 = im2[tly:tly+dely, tlx:tlx+delx]
        # rescale image to deal with variation in lighting
        im4 = exposure.rescale_intensity(im3)
        # make binary image
        val = filter.threshold_otsu(im4)
        im5 = im4 > val
        im5.astype(int)
        # calculate the area of normally dark ground that became light
        snow_max = sum_snow - sum_nosnow
        img_sum = im5.sum()
        pix_diff = sum_snow - img_sum
        snow_pix = snow_max - pix_diff
        snow_cover_percent = int(snow_pix / (1.0 * snow_max) * 100)
        # stash result
        depth_list.append(snow_cover_percent)
    # turn lists to numpy arrays
    depth_list = np.array((depth_list))
    time_list = np.array((time_list))
    # calculate within replicate means and standard deviations
    starts = range(0, len(depth_list), rep_size)
    rep_means = []
    rep_std = []
    rep_std_flag = []
    rep_time = []
    for start in starts:
        # create start and end indices
        block_start = start
        block_end = start + rep_size
        # get mean for block
        mean_i = np.nanmean(depth_list[block_start : block_end])
        # get sd for block
        sd_i = np.nanstd(depth_list[block_start : block_end])
        # set quality flag
        sd_flag_i = 0
        if sd_i < sd_thresh:
            sd_flag_i = 0
        else:
            sd_flag_i = 1
        # get mean time of block
        med_index = block_start
        time_i = time_list[med_index]
        # append to lists
        rep_means.append(mean_i)
        rep_std.append(sd_i)
        rep_std_flag.append(sd_flag_i)
        rep_time.append(time_i)
    out_list = [rep_means, rep_std, rep_std_flag, rep_time]
    #depth_lst = pd.rolling_median(depth_lst, window=3, min_periods=1)
    return out_list

# produce summary data frame
def summarize_output(all_output_list, depth = 1):
    """
    sumarize output from depth and cover functions and outputs spreadsheet-
    like data.
    """
    all_depths_list = all_output_list
    # pull out the times
    t_depths = all_depths_list[0][3]
    # pull out the depths
    depths = []
    for i in range(0,len(all_depths_list)):
            depths_i = all_depths_list[i][0]
            print depths_i
            depths.append(depths_i)
    depths = np.array(depths)
    # pull out the flags
    flags = []
    for i in range(0,len(all_depths_list)):
            flags_i = all_depths_list[i][2]
            print flags_i
            flags.append(flags_i)
    flags = np.array(flags)
    # make nan any depth with an sd flag
    for i in range(0,len(flags)):
        for j in range(len(flags[0])):
            fi = flags[i,j]
            if fi == 1:
                depths[i,j] = np.nan
    # get average, n, se of depths
    mean_depths = []
    se_depths = []
    n_depths = []
    for i in range(0,len(depths[0])):
        mean_i = np.nanmean(depths[:,i])
        sd_i = np.nanstd(depths[:,i])
        n_i = 3 - np.count_nonzero(np.isnan(depths[:,i]))
        se_i = sd_i / (1.0 * np.sqrt(n_i))
        mean_depths.append(mean_i)
        se_depths.append(se_i)
        n_depths.append(n_i)
    mean_depths = np.array(mean_depths)
    se_depths = np.array(se_depths)
    n_depths = np.array(n_depths)
    # assemble
    out_array = np.array([t_depths, mean_depths, se_depths, n_depths]).T
    out_df = pd.DataFrame(out_array)
    # give the data frame some column names
    if depth == 1:
        out_df.columns = ["time", "cmMeanDepth", "cmSEMeanDepth", "nMeanDepth"]
    else:
        out_df.columns = ["time", "meanPercentCover", "sePercentCover", 
        "nPercentCover"]
    return out_df


##########################  get input images  #################################

"""
Next we decide on a collection of images to analyze.  In this case I am using
images with time stamps beginning with 07 hours (7 am, dawn) and 17 (5 pm,
dusk).  Dawn and dusk images are used because direct light on the snow gauges 
looks a lot like snow, and confounds measurments.
"""

# set directory with images
work_dir = "C:\\Users\\tmeehan\\Documents\\Work\\NEONWork\\\
ImageProcessing\\Snow\\images2"
os.chdir(work_dir)

# grab files
rep_size = 4
f1 = glob.glob('*_07????.jpg')
f11 = glob.glob('*_17????.jpg')
flist = f1+f11
flist = sorted(flist,key=os.path.getmtime); flist


######################  get depths and percent covers  ########################

"""
Now we run the depth function for 3 snow gauges and the cover function for
3 ground reference areas.  The hardest part of running these functions, by
far, is getting all the right parameter values. It is beyond the scope of
this example to describe this process, though.
"""

# get depth for center ruler
depth_list1 = get_ruler_depths(flist = flist, rotangle = -2.0, tlx = 575,
                               tly = 445, delx = 12, dely = 472, x0 = 9, 
                               y0 = 1, x1 = 5, y1 = 470, cm = 100.0, 
                               pix = 219.0, cut = 9, rep_size = 4, 
                               sd_thresh = 10)

# get depth for left-most ruler
depth_list2 = get_ruler_depths(flist = flist, rotangle = -5.0, tlx = 489,
                               tly = 90, delx = 9, dely = 310, x0 = 4, 
                               y0 = 4, x1 = 4, y1 = 306, cm = 100.0, 
                               pix = 151.0, cut = 9, rep_size = 4, 
                               sd_thresh = 10)

# get depth for right-most ruler
depth_list3 = get_ruler_depths(flist = flist, rotangle = 9.0, tlx = 1123,
                               tly = 110, delx = 10, dely = 309, x0 = 4, 
                               y0 = 3, x1 = 4, y1 = 307, cm = 100.0, 
                               pix = 151.0, cut = 9, rep_size = 4, 
                               sd_thresh = 10)

# get percent snow cover from first reference patch
cover_list1 = get_snow_cover(flist, tlx = 234, tly = 63, delx = 129,
                             dely = 53, sum_snow = 5180, sum_nosnow = 2536)

# get percent snow cover from second reference patch
cover_list2 = get_snow_cover(flist, tlx = 821, tly = 4, delx = 148,
                             dely = 58, sum_snow = 5667, sum_nosnow = 3272)

# get percent snow cover from second reference patch
cover_list3 = get_snow_cover(flist, tlx = 535, tly = 386, delx = 186,
                             dely = 80, sum_snow = 13422, sum_nosnow = 5812)


###############################  summarize results  ###########################

"""
Now we can run the summary function on the results from the depth and cover
function.  This will put a copy in memory and on disk in the working directory.
"""

# gather results
all_depths_list = [depth_list1, depth_list2, depth_list3]
all_covers_list = [cover_list1, cover_list2, cover_list3]

# run summary function
snow_depth_data = summarize_output(all_output_list = all_depths_list, depth = 1)
snow_cover_data = summarize_output(all_output_list = all_covers_list, depth = 0)

# send to CSVs
snow_depth_data.to_csv("snow_depth_data.csv", index=False)
snow_cover_data.to_csv("snow_cover_data.csv", index=False)


##################################  plot results  #############################

"""
Now let's make some time series plots of snow depth and cover, and add some
images from the snow gauge camera.
"""

# set window option
%matplotlib qt

# make plot
fig = plt.figure()
gs = gridspec.GridSpec(nrows=3, ncols=2)

# add subplots
ax1 = plt.subplot(gs[0,0])
ax1.xaxis.set_major_formatter(mpd.DateFormatter('%d'))
ax1.plot_date(snow_depth_data["time"], snow_depth_data["cmMeanDepth"],
              ls='-', marker='o')
ax1.set_ylabel('Snow depth (mean, SE, cm)')
ax1.set_xlabel('Sample date (Feb 2016)')
ax1.grid(False)
ax1.axhline(0)
ax1.errorbar(snow_depth_data["time"], snow_depth_data["cmMeanDepth"],
             yerr=snow_depth_data["cmSEMeanDepth"], fmt='-o')

ax2 = plt.subplot(gs[0,1])
ax2.xaxis.set_major_formatter(mpd.DateFormatter('%d'))
ax2.plot_date(snow_cover_data["time"], snow_cover_data["meanPercentCover"],
              ls='-', marker='o', color="red")
ax2.set_ylabel('Snow cover (mean, SE, %)')
ax2.set_xlabel('Sample date (Feb 2016)')
ax2.grid(False)
ax2.axhline(0)
ax2.errorbar(snow_cover_data["time"], snow_cover_data["meanPercentCover"],
             yerr=snow_cover_data["sePercentCover"], fmt='-o', color="red")

ax3 = plt.subplot(gs[1,0])
image3 = ax3.imshow(nd.imread(flist[6], flatten=0), aspect='auto')
#plt.setp(ax3.get_yticklabels(), visible=False)
plt.setp(ax3, title="13 Feb 2016")

ax4 = plt.subplot(gs[1,1])
image4 = ax4.imshow(nd.imread(flist[25], flatten=0), aspect='auto')
plt.setp(ax4, title="15 Feb 2016")

ax5 = plt.subplot(gs[2,0])
image5 = ax5.imshow(nd.imread(flist[32], flatten=0), aspect='auto')
plt.setp(ax5, title="17 Feb 2016")

ax6 = plt.subplot(gs[2,1])
image6 = ax6.imshow(nd.imread(flist[55], flatten=0), aspect='auto')
plt.setp(ax6, title="19 Feb 2016")

# plot
gs.tight_layout(fig)
plt.show()

# save and show figure
savefig("Snow_time_series_plot.pdf", dp1=300, format="pdf")


####################################  make animation  #########################

"""
You can also create an animation from the list of images that were just 
processed, to see if the output data is reasonable.
"""

# set window option
%matplotlib qt

# collect images
imagelist = [misc.imread(i, flatten=0) for i in flist]
fig = plt.figure() # make figure

# make axesimage object
im = plt.imshow(imagelist[0], cmap=plt.get_cmap('jet'), vmin=0, vmax=255)

# function to update figure
def updatefig(j):
    # set the data in the axesimage object
    im.set_array(imagelist[j])
    # return the artists set
    return im,
# kick off the animation
ani = animation.FuncAnimation(fig, updatefig, frames=len(flist),
                              interval=800, blit=True)

